@@ -1,35 +1,33 @@
+import itertools
 import os
 import random
-import typing
-import numpy as np
-import itertools
 
+import numpy as np
 import pytest
-from emmet.core.summary import HasProps
-from emmet.core.symmetry import CrystalSystem
 from emmet.core.tasks import TaskDoc
 from emmet.core.vasp.calc_types import CalcType
-from sympy import prime
-from mp_api.client.core.settings import MAPIClientSettings
-from mp_api.client import MPRester
-from pymatgen.analysis.magnetism import Ordering
 from pymatgen.analysis.phase_diagram import PhaseDiagram
 from pymatgen.analysis.pourbaix_diagram import IonEntry, PourbaixDiagram, PourbaixEntry
 from pymatgen.analysis.wulff import WulffShape
-from pymatgen.core.periodic_table import Element
 from pymatgen.core.ion import Ion
+from pymatgen.core.periodic_table import Element
 from pymatgen.electronic_structure.bandstructure import (
     BandStructure,
     BandStructureSymmLine,
 )
 from pymatgen.electronic_structure.dos import CompleteDos
-from pymatgen.entries.computed_entries import ComputedEntry, GibbsComputedStructureEntry
 from pymatgen.entries.compatibility import MaterialsProjectAqueousCompatibility
+from pymatgen.entries.computed_entries import ComputedEntry, GibbsComputedStructureEntry
 from pymatgen.io.cif import CifParser
 from pymatgen.io.vasp import Chgcar
 from pymatgen.phonon.bandstructure import PhononBandStructureSymmLine
 from pymatgen.phonon.dos import PhononDos
 
+from mp_api.client import MPRester
+from mp_api.client.core.settings import MAPIClientSettings
+
+os.environ["MP_API_KEY"] = "test"
+
 
 @pytest.fixture()
 def mpr():
@@ -38,9 +36,7 @@ def mpr():
     rester.session.close()
 
 
-@pytest.mark.skipif(
-    os.environ.get("MP_API_KEY", None) is None, reason="No API key found."
-)
+@pytest.mark.skipif(os.getenv("MP_API_KEY", None) is None, reason="No API key found.")
 class TestMPRester:
     def test_get_structure_by_material_id(self, mpr):
         s1 = mpr.get_structure_by_material_id("mp-149")
@@ -50,7 +46,7 @@ class TestMPRester:
         assert s1.formula == "Ca12 Ti8 O28"
 
         s1 = mpr.get_structure_by_material_id("mp-149", final=False)
-        assert [s.formula for s in s1] == ["Si2"]
+        assert {s.formula for s in s1} == {"Si2"}
 
         # # requesting via task-id instead of mp-id
         with pytest.warns(UserWarning):
@@ -74,14 +70,14 @@ class TestMPRester:
         assert len(data) > 5
 
     def test_get_materials_ids_doc(self, mpr):
-        mpids = mpr.get_materials_ids("Al2O3")
-        random.shuffle(mpids)
-        doc = mpr.materials.get_data_by_id(mpids.pop(0))
+        mp_ids = mpr.get_materials_ids("Al2O3")
+        random.shuffle(mp_ids)
+        doc = mpr.materials.get_data_by_id(mp_ids.pop(0))
         assert doc.formula_pretty == "Al2O3"
 
-        mpids = mpr.get_materials_ids("Al-O")
-        random.shuffle(mpids)
-        doc = mpr.materials.get_data_by_id(mpids.pop(0))
+        mp_ids = mpr.get_materials_ids("Al-O")
+        random.shuffle(mp_ids)
+        doc = mpr.materials.get_data_by_id(mp_ids.pop(0))
         assert doc.chemsys == "Al-O"
 
     def test_get_structures(self, mpr):
@@ -105,9 +101,9 @@ class TestMPRester:
     def test_get_bandstructure_by_material_id(self, mpr):
         bs = mpr.get_bandstructure_by_material_id("mp-149")
         assert isinstance(bs, BandStructureSymmLine)
-        bs_unif = mpr.get_bandstructure_by_material_id("mp-149", line_mode=False)
-        assert isinstance(bs_unif, BandStructure)
-        assert not isinstance(bs_unif, BandStructureSymmLine)
+        bs_uniform = mpr.get_bandstructure_by_material_id("mp-149", line_mode=False)
+        assert isinstance(bs_uniform, BandStructure)
+        assert not isinstance(bs_uniform, BandStructureSymmLine)
 
     def test_get_dos_by_id(self, mpr):
         dos = mpr.get_dos_by_material_id("mp-149")
@@ -229,7 +225,7 @@ class TestMPRester:
         # test removal of extra elements from reference solids
         # Li-Zn-S has Na in reference solids
         pbx_entries = mpr.get_pourbaix_entries("Li-Zn-S")
-        assert not any([e for e in pbx_entries if "Na" in e.composition])
+        assert not any(e for e in pbx_entries if "Na" in e.composition)
 
         # Ensure entries are pourbaix compatible
         PourbaixDiagram(pbx_entries)
@@ -254,6 +250,10 @@ class TestMPRester:
         ion_entries = mpr.get_ion_entries(pd, ion_entry_data)
         assert len(ion_entries) == 5
         assert all([isinstance(i, IonEntry) for i in ion_entries])
+        bi_v_entry_data = mpr.get_ion_reference_data_for_chemsys("Bi-V")
+        bi_data = mpr.get_ion_reference_data_for_chemsys("Bi")
+        v_data = mpr.get_ion_reference_data_for_chemsys("V")
+        assert len(bi_v_entry_data) == len(bi_data) + v_data
 
         # test an incomplete phase diagram
         entries = mpr.get_entries_in_chemsys("Ti-O")
@@ -270,7 +270,7 @@ class TestMPRester:
             itertools.chain.from_iterable(i.elements for i in ion_ref_comps)
         )
         ion_ref_entries = mpr.get_entries_in_chemsys(
-            list([str(e) for e in ion_ref_elts] + ["O", "H"])
+            [*map(str, ion_ref_elts), "O", "H"]
         )
         mpc = MaterialsProjectAqueousCompatibility()
         ion_ref_entries = mpc.process_entries(ion_ref_entries)
@@ -294,13 +294,13 @@ class TestMPRester:
         assert np.allclose(so4_two_minus.energy, solid_energy + 5.542, atol=1e-3)
 
     def test_get_phonon_data_by_material_id(self, mpr):
-        bs = mpr.get_phonon_bandstructure_by_material_id("mp-11659")
+        bs = mpr.get_phonon_bandstructure_by_material_id("mp-2172")
         assert isinstance(bs, PhononBandStructureSymmLine)
 
-        dos = mpr.get_phonon_dos_by_material_id("mp-11659")
+        dos = mpr.get_phonon_dos_by_material_id("mp-2172")
         assert isinstance(dos, PhononDos)
 
-    @pytest.mark.skip(reason="Test needs fixing with ENV variables")
+    # @pytest.mark.skip(reason="Test needs fixing with ENV variables")
     def test_get_charge_density_data(self, mpr):
         chgcar = mpr.get_charge_density_from_material_id("mp-149")
         assert isinstance(chgcar, Chgcar)
@@ -324,3 +324,17 @@ class TestMPRester:
         ]
         docs = mpr.summary.search(material_ids=mpids, fields=["material_ids"])
         assert len(docs) == 15000
+
+
+def test_pmg_api_key(monkeypatch: pytest.MonkeyPatch):
+    from pymatgen.core import SETTINGS
+
+    # unset DEFAULT_API_KEY
+    monkeypatch.setattr("mp_api.client.mprester.DEFAULT_API_KEY", None)
+
+    fake_api_key = "12345678901234567890123456789012"  # 32 chars
+    # patch pymatgen.core.SETTINGS to contain PMG_MAPI_KEY
+    monkeypatch.setitem(SETTINGS, "PMG_MAPI_KEY", fake_api_key)
+
+    # create MPRester and check that it picked up the API key from pymatgen SETTINGS
+    assert MPRester().api_key == fake_api_key
